@@ -3,7 +3,6 @@ import os, re
 import json
 import time
 import signal
-import atexit
 import loggus
 import socket
 import threading
@@ -11,40 +10,31 @@ import selectors
 
 from _io import BufferedReader
 from datetime import timedelta
+from collections import defaultdict
 from pywss.headers import *
 from pywss.statuscode import *
-from pywss.websocket import WebsocketContextWrap
+from pywss.websocket import WebSocketContextWrap
 from pywss.static import NewStaticHandler
+from pywss.routing import Route
+from pywss.openapi import openapi_ui_template, merge_dict, parameters_filter
 
-__version__ = '0.1.2'
-
-
-class _Run:
-    running = True
-
-    @classmethod
-    def stop(cls):
-        cls.running = False
-
-
-signal.signal(signal.SIGTERM, lambda *args: _Run.stop())
-signal.signal(signal.SIGINT, lambda *args: _Run.stop())
-signal.signal(signal.SIGILL, lambda *args: _Run.stop())
-atexit.register(lambda: _Run.stop())
+__version__ = '0.1.3'
 
 
 class Context:
     _handler_index = 0
     _flush_header = False
 
-    def __init__(self, fd, r, method, path, version, headers, route, handlers, address):
+    def __init__(self, app, fd, rfd, method, path, paths, version, headers, route, handlers, address):
+        self.app = app
         self.fd = fd
-        self.rfd = r
+        self.rfd = rfd
         self.method = method
         self.version = version
         self.headers = headers
         self.cookies = parse_cookies(headers)
         self.path = path
+        self.paths = paths
         self.params = parse_params(path)
         self.route = route
         self._handlers = handlers
@@ -54,7 +44,7 @@ class Context:
         self.content_length = int(headers.get("Content-Length", 0))
         self.content = b""
         if self.content_length:
-            self.content = r.read(self.content_length)
+            self.content = rfd.read(self.content_length)
 
         self.response_status_code = 200
         self.response_headers = {
@@ -239,23 +229,60 @@ class Context:
 class App:
 
     def __init__(self, base_route="", base_handlers=None):
+        self.running = True
         self.base_route = f"/{base_route.strip().strip('/')}" if base_route else base_route
         self.base_handlers = list(base_handlers) if base_handlers else []
         self.head_match_routes = []
+        self.parse_match_routes = []
         self.full_match_routes = {}
         self.log = loggus.GetLogger()
+        self.openapi_data = {
+            "paths": defaultdict(dict),
+        }
 
-    def register(self, method, route, handlers):
+    def register(self, method, route, handlers) -> None:
         route = f"/{route.strip().strip('/')}" if route else route
         self.full_match_routes[f"{method}{self.base_route}{route}"] = list(self.base_handlers) + list(handlers)
 
-    def build(self):
+    def build(self) -> None:
         routes = {}
         for route, v in self.full_match_routes.items():
             if isinstance(v, App):
                 v.build()
                 routes.update(v.full_match_routes)
                 self.head_match_routes += v.head_match_routes
+                self.parse_match_routes += v.parse_match_routes
+                self.openapi_data = merge_dict(self.openapi_data, v.openapi_data)
+        for route, v in self.full_match_routes.items():
+            if isinstance(v, App):
+                continue
+            elif "{" in route and "}" in route:
+                r = Route.from_route(route)
+                self.parse_match_routes.append(
+                    (r, v)
+                )
+                if hasattr(v[-1], "__openapi_path__"):
+                    i = route.index("/")
+                    _method, _route = route[:i], route[i:]
+                    path = v[-1].__openapi_path__
+                    for node in r.route_list:
+                        if not node.name:
+                            continue
+                        if parameters_filter(node.name, path["parameters"]):
+                            continue
+                        path["parameters"].append({
+                            "name": node.name,
+                            "in": "path",
+                            "required": True,
+                        })
+                    if _route not in self.openapi_data["paths"]:
+                        self.openapi_data["paths"][_route] = {}
+                    self.openapi_data["paths"][_route][_method.lower()] = path
+                self.log.update({
+                    "type": "parsermatch",
+                    "route": route,
+                    "handlers": [handler.__name__ for handler in v],
+                }).info(f"bind route")
             elif route.endswith("*"):
                 match = route.strip().rstrip("/*").rstrip("*")
                 self.head_match_routes.append((match, v))
@@ -266,6 +293,12 @@ class App:
                 }).info(f"bind route")
             else:
                 routes[route] = v
+                if hasattr(v[-1], "__openapi_path__"):
+                    i = route.index("/")
+                    _method, _route = route[:i], route[i:]
+                    if _route not in self.openapi_data["paths"]:
+                        self.openapi_data["paths"][_route] = {}
+                    self.openapi_data["paths"][_route][_method.lower()] = v[-1].__openapi_path__
                 self.log.update({
                     "type": "fullmatch",
                     "route": route,
@@ -273,15 +306,16 @@ class App:
                 }).info(f"bind route")
         self.full_match_routes = routes
 
-    def party(self, route, *handlers):
+    def party(self, route, *handlers) -> 'App':
         route = f"{self.base_route}/{route.strip().strip('/')}"
-        self.full_match_routes[f"{route}"] = App(route, list(self.base_handlers) + list(handlers))
-        return self.full_match_routes[f"{route}"]
+        if route not in self.full_match_routes:
+            self.full_match_routes[route] = App(route, list(self.base_handlers) + list(handlers))
+        return self.full_match_routes[route]
 
-    def use(self, *handlers):
+    def use(self, *handlers) -> None:
         self.base_handlers += list(handlers)
 
-    def static(self, route, rootDir, *handlers, staticHandler=NewStaticHandler):
+    def static(self, route, rootDir, *handlers, staticHandler=NewStaticHandler) -> None:
         if staticHandler:
             handlers = list(handlers)
             handlers.append(staticHandler(rootDir))
@@ -289,28 +323,28 @@ class App:
             route = f"{route.strip().rstrip('/')}/*"
         self.register("GET", route, handlers)
 
-    def get(self, route, *handlers):
+    def get(self, route, *handlers) -> None:
         self.register("GET", route, handlers)
 
-    def post(self, route, *handlers):
+    def post(self, route, *handlers) -> None:
         self.register("POST", route, handlers)
 
-    def head(self, route, *handlers):
+    def head(self, route, *handlers) -> None:
         self.register("HEAD", route, handlers)
 
-    def put(self, route, *handlers):
+    def put(self, route, *handlers) -> None:
         self.register("PUT", route, handlers)
 
-    def delete(self, route, *handlers):
+    def delete(self, route, *handlers) -> None:
         self.register("DELETE", route, handlers)
 
-    def patch(self, route, *handlers):
+    def patch(self, route, *handlers) -> None:
         self.register("PATCH", route, handlers)
 
-    def options(self, route, *handlers):
+    def options(self, route, *handlers) -> None:
         self.register("OPTIONS", route, handlers)
 
-    def any(self, route, *handlers):
+    def any(self, route, *handlers) -> None:
         self.get(route, *handlers)
         self.post(route, *handlers)
         self.head(route, *handlers)
@@ -319,34 +353,70 @@ class App:
         self.patch(route, *handlers)
         self.options(route, *handlers)
 
-    def _(self, request, address):
+    def openapi(
+            self,
+            title="OpenAPI",
+            version="0.0.1",
+            openapi_json_route="/openapi.json",
+            openapi_ui_route="/docs",
+            openapi_ui_js_url="https://cdn.bootcdn.net/ajax/libs/swagger-ui/4.14.0/swagger-ui-bundle.js",
+            openapi_ui_css_url="https://cdn.bootcdn.net/ajax/libs/swagger-ui/4.14.0/swagger-ui.css",
+    ) -> None:
+        self.openapi_data = {
+            "openapi": "3.0.2",
+            "info": {
+                "title": title,
+                "version": version
+            },
+            "paths": defaultdict(dict),
+        }
+        self.get(openapi_json_route, lambda ctx: ctx.write(self.openapi_data))
+        self.get(openapi_ui_route, lambda ctx: ctx.write(openapi_ui_template(
+            title,
+            f"{self.base_route}/{openapi_json_route.strip('/')}",
+            openapi_ui_js_url,
+            openapi_ui_css_url,
+        )))
+
+    def _(self, request, address) -> None:
         log = self.log
         try:
-            r = request.makefile("rb", -1)
-            method, path, version, err = parse_request_line(r)
+            rfd = request.makefile("rb", -1)
+            method, path, version, err = parse_request_line(rfd)
             if err:
-                request.sendall(b"HTTP/1.1 404 Pywss\r\n")
+                request.sendall(b"HTTP/1.1 400 BadRequest\r\n")
                 return
             log = log.update(method=method, path=path)
-            hes, err = parse_headers(r)
+            hes, err = parse_headers(rfd)
             if err:
-                request.sendall(b"HTTP/1.1 404 Pywss\r\n")
+                request.sendall(b"HTTP/1.1 400 BadRequest\r\n")
                 log.error(err)
                 return
             # full match
+            paths = {}
             route = f"{method.upper()}/{path.split('?', 1)[0].strip('/')}"
             handlers = self.full_match_routes.get(route, None)
+            # parser match
+            if not handlers:
+                for r, v in self.parse_match_routes:
+                    fix, ps = r.match(route)
+                    if fix:
+                        route = r.route
+                        handlers = v
+                        paths = ps
+                        break
             # head match
             if not handlers:
                 for r, v in self.head_match_routes:
                     if route.startswith(r):
                         route = r
                         handlers = v
+                        break
             if not handlers:
-                request.sendall(b"HTTP/1.1 404 Pywss\r\n")
-                log.error("No Handler")
+                request.sendall(b"HTTP/1.1 404 NotFound\r\n")
+                log.warning("No Handler")
                 return
-            ctx = Context(request, r, method, path, version, hes, route, handlers, address)
+            ctx = Context(self, request, rfd, method, path, paths, version, hes, route, handlers, address)
             ctx.log = log
             ctx.next()
             ctx.flush()
@@ -355,7 +425,14 @@ class App:
         finally:
             request.close()
 
-    def run(self, host="0.0.0.0", port=8080, grace=3, request_queue_size=5, poll_interval=0.5):
+    def close(self) -> None:
+        self.running = False
+
+    def run(self, host="0.0.0.0", port=8080, grace=0, request_queue_size=5, poll_interval=0.5) -> None:
+        signal.signal(signal.SIGTERM, lambda *args: self.close())
+        signal.signal(signal.SIGINT, lambda *args: self.close())
+        signal.signal(signal.SIGILL, lambda *args: self.close())
+
         self.build()
         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
         sock.bind((host, port))
@@ -368,11 +445,13 @@ class App:
                     update(version=__version__). \
                     variables(host, port, grace). \
                     info("server start")
-                while _Run.running:
+                while self.running:
                     ready = _selector.select(poll_interval)
                     if not ready:
                         continue
                     request, address = sock.accept()
                     threading.Thread(target=self._, args=(request, address)).start()
-        time.sleep(grace)
+        for i in range(int(grace) + 1):
+            self.log.update(hit=i + 1, grace=grace).warning("server closing")
+            time.sleep(1)
         self.log.panic("server closed")
